/*
 * David Wehr http://davidawehr.com/
 * Creative Commons Attribution 3.0 Unported (CC BY 3.0)
 * http://creativecommons.org/licenses/by/3.0/
 */

import processing.core.PApplet;
import processing.core.PImage;

public class Geotweet_mapper extends PApplet{
	Data d; //Storage container for global data
	
	//Object for dynamic buttons or labels
	public class Button {
		
		private int x_start;
		private int y_start;
		private int x_end;
		private int y_end;
		
		//Used for making button object but modifying later
		public Button() {
			//An empty constructor
			x_start = 0;
			y_start = 0;
			x_end = 0;
			y_end = 0;
		}
		
		//Standard button creator. Draws button with dimensions
		public Button(String label, int x, int y, int w, int h, int r) {
			/* Parameters,
			 * label - What the text will display
			 * x - x coordinate for position of upper left corner(pixels)
			 * y - y coordinate for position of upper left corner(pixels)
			 * w - width of button(pixels)
			 * h - height of button(pixels)
			 * r - corner radius(pixels)
			 */
			draw(label, x, y, w, h, r);
		}
		
		//Can be called anytime after button creation to modify the button's properties
		public void draw(String label, int x, int y, int w, int h, int r) {
			//Same parameters as constructor
			stroke(1);
			fill(1);
			rect(x_start-1, y_start-1, x_end-x_start+2, y_end-y_start+2);
			
			x_start = x;
			y_start = y;
			x_end = x+w;
			y_end = y+h;
			stroke(0);
			fill(0.9f);
			rect(x, y, w, h, r);
			
			fill(0);
			textAlign(LEFT);
			textSize(y_end - y_start - 3);
			text(label, x_start+3, y_end-5);
		}
		
		//Used to check of the button has been clicked.
		public boolean mouseOver() {
			if (mouseX >= x_start && mouseX <= x_end &&
					mouseY >= y_start && mouseY <= y_end) {
				return true;
			}
			else {
				return false;
			}
		}
	}
	private int[] convertToScreen(double longitude, double latitude) {
		//Converts earth longitude and latitude coordinates(in degrees) to
		//their corresponding point on a mollweide projection.

		longitude = longitude / 57.296;
		latitude = latitude / 57.296;

		double dtheta = 0.149;
		double theta = latitude;
		double pi_sin_lat = 3.1416 * Math.sin(latitude);
		//Uses newton's method to calculate theta to reasonable accuracy
		int i = 0;
		while (Math.abs (dtheta) > 0.00000005) {
			if (i == 30) {
				break;
			}
			dtheta = -(theta + Math.sin(theta) - pi_sin_lat) / (1 + Math.cos(theta));
	 		theta += dtheta;
	 		i += 1;
		}
		
		theta = theta / 2;
		double x = (longitude * Math.cos(theta)) / 3.1416;
		double y = Math.sin(theta);


		int[] coords = new int[2];

		coords[0] = (int) (((x + 1)/2) * (d.sizeX-100));// + 100;
		coords[1] = (int) (((-y + 1)/2) * (d.sizeY-50));// + 50;

		return coords;
	}
	

	private void markPoint(double longitude, double latitude) {
		//Handles the graphics of marking points on the map
		int[] place_coords = convertToScreen(longitude, latitude);
		int centerX = place_coords[0];
		int centerY = place_coords[1];
		
		
		heat_map.loadPixels();
		heat_img.loadPixels();
		//Iterate through each pixel in heat_img(A marker dot) and overlays the proper color and alpha
		//value depending on the strength.
		for (int r = 0; r <= 11; r++) {
			int y = centerY + r - 6;
			for (int c = 0; c <= 11; c++) {
				float strength =  brightness(heat_img.pixels[r*12 + c]) * 0.35f;
				int x = centerX + c - 6;
				
				float old_strength = (0.6667f - hue(heat_map.pixels[y*(d.sizeX-100) + x])) / 0.6667f;
				float new_hue = 0.6667f - (old_strength + strength) * 0.6667f;
				int new_color = color(new_hue, 1, 1, (old_strength+strength)*3f);
				
				heat_map.pixels[y*(d.sizeX-100) + x] = new_color;
			}
		}
		heat_map.updatePixels();
		//Place world map on 
		image(world_map, 100, 50, d.sizeX-100, d.sizeY-50);
		image(heat_map, 100, 50, d.sizeX-100, d.sizeY-50);
	}
	
	
	private void makeUI() {
		//Draws the user interface
		//Can be used to clear the heatmap
		
		background(1);
		
		image(world_map, 100, 50, d.sizeX-100, d.sizeY-50);
		
		//Makes an empty image for the heatmap overlay 
		heat_map = createImage(d.sizeX-100, d.sizeY-50, ARGB);
		
		//Make the pixels blue for the lowest strength, but transparent so it's not visible
		heat_map.loadPixels();
		int blank = color(0.6667f, 1, 1, 0);
		for (int i = 0; i < heat_map.pixels.length; i++) {
			heat_map.pixels[i] = blank;
		}
		heat_map.updatePixels();
		
		
		//Draw buttons to match the current state
		if (!d.streaming) {
			d.toggle_connection.draw("Start streaming", 3, 5, 131, 20, 5);
		}
		
		if (d.streaming) {
			d.toggle_connection.draw("Stop streaming", 3, 5, 128, 20, 5);
		}
		
		clear_screen.draw("Clear", 3, 30, 48, 20, 5);
		
		if (d.stream_delay == 20) {
			change_rate.draw("Maximum Speed", 3, 55, 143, 20, 5);
		}
		if (d.stream_delay == 60) {
			change_rate.draw("Medium Speed", 3, 55, 128, 20, 5);
		}
		if (d.stream_delay == 100) {
			change_rate.draw("Minimum Speed", 3, 55, 141, 20, 5);
		}
		
		rate_display.draw("-- tweets/sec", d.sizeX - 135, 5, 130, 20, 2);
		
	}
	
	void startConnection() {
		//The TweetReader thread checks this value to see if it needs to connect
		d.start_stream = true;
		
		d.toggle_connection.draw("Connecting...", 3, 5, 116, 20, 5);
	}
	
	PImage heat_map;
	PImage heat_img;
	PImage world_map;
	
	Button clear_screen;
	Button rate_display;
	Button change_rate;
	public void setup() {
		//Runs once at the start of the application
		
		d = Data.getData();
		
		d.tweetReader = new TweetReader();
		
		//To use, put your own twitter OAuth information here. Read-only access level is fine.
		d.con_key = "Consumer Key";
		d.con_sec = "Consumer Secret";
		d.acc_token = "Access Token";
		d.acc_token_sec = "Access token secret";
		
		//All geocoded tweets around the whole world
		d.url = "https://stream.twitter.com/1.1/statuses/filter.json?locations=-180%2C-90%2C180%2C90";
		
		d.sizeX = 800;
		d.sizeY = d.sizeX/2;
		
		d.stream_delay = 100;
		
		colorMode(HSB, 1.0f);
//		frame.setResizable(true); //Needed if running in Processing IDE
		size(d.sizeX, d.sizeY);
		world_map = loadImage("world_map_mollweide.png");
		heat_img = loadImage("heat.png");
		
		d.toggle_connection = new Button();
		clear_screen = new Button();
		rate_display = new Button();
		change_rate = new Button();
		
		makeUI();
	}
	
	int rate_tweets = 0;
	int rate_start = 0;
	public void draw() {
		if (d.streaming && d.changed) {
			markPoint(d.last_tweet[0], d.last_tweet[1]);
			rate_tweets ++;
			d.changed = false;
		}
		
		//Every five seconds, calculate the average tweet rate based on the past 5 seconds of data
		if ((millis() - rate_start) > 5000 && d.streaming) {
			int rate = rate_tweets / ((millis() - rate_start)/1000);
			rate_display.draw(Integer.toString(rate) + " tweets/sec", d.sizeX - 135, 5, 130, 20, 2);
			rate_start = millis();
			rate_tweets = 0;
		}
		
		//Check to see if the window was resized. Some sort of callback system would better suit this
		if (width != d.sizeX) {
			height = width / 2;
			d.sizeX = width;
			d.sizeY = height;
			makeUI();
		}
	}
	
	public void mouseClicked() {
		//This function from the processing library runs when the mouse is clicked
		//Check if the mouse is over any buttons, and run the corresponding action
		
		if (d.toggle_connection.mouseOver() &&  !d.streaming) {
			startConnection();
			
		}
		else if (d.toggle_connection.mouseOver() && d.streaming) {
			d.kill_stream = true;
			d.streaming = false;
			d.toggle_connection.draw("Start streaming", 3, 5, 131, 20, 5);
		}
		
		else if (change_rate.mouseOver()) {
			if (d.stream_delay == 100) {
				d.stream_delay = 60;
				change_rate.draw("Medium Speed", 3, 55, 128, 20, 5);
			}
			else if (d.stream_delay == 60) {
				d.stream_delay = 20;
				change_rate.draw("Maximum Speed", 3, 55, 143, 20, 5);
			}
			else if (d.stream_delay == 20) {
				d.stream_delay = 100;
				change_rate.draw("Minimum Speed", 3, 55, 141, 20, 5);
			}
		}
		
		else if (clear_screen.mouseOver()) {
			makeUI();
		}
	}
	
	public void keyPressed() {
		if (key == 27) {
			//esc key. Tries to close the stream properly, but will exit anyway.
			d.kill_stream = true;
			exit();
		}
	}
}
