import java.io.IOException;

import java.io.InputStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.scribe.builder.ServiceBuilder;
import org.scribe.builder.api.TwitterApi;
import org.scribe.exceptions.OAuthException;
import org.scribe.model.OAuthRequest;
import org.scribe.model.Response;
import org.scribe.model.Token;
import org.scribe.model.Verb;
import org.scribe.oauth.OAuthService;

//Handles the actual connection to the stream. Runs in a different thread to prevent
//the main application from hanging when the connection is lost
public class TweetReader implements Runnable{
	public InputStream stream;
	public InputStreamReader inReader;
	public BufferedReader buffReader;
	
	//Data object to be accessed throughout the class
	private Data d;
	
	public TweetReader() {
		d = Data.getData();
		Thread t = new Thread(this);
		t.start();
	}
	
	public void run() {
		//Constantly check if the stream needs started, killed, or if currently streaming
		while (true) {
			//Opens the connection
			if (d.start_stream) {
				connect(
					d.url, 
					d.con_key, 
					d.con_sec, 
					d.acc_token, 
					d.acc_token_sec);

				d.start_stream = false;
				d.toggle_connection.draw("Stop streaming", 3, 5, 128, 20, 5);
			}
			
			//Closes the connection
			if (d.kill_stream) {
				try {
					buffReader.close();
					d.streaming = false;
					d.kill_stream = false;
				} catch (IOException ie){}
			}
			
			//Reads tweets
			if (d.streaming){
				try {
					getTweet();
				} catch(Exception e) {
					//System.out.println(e.getMessage());
				}
			}
			
			//Sleeps the thread depending on the speed of the stream
			try {
				Thread.sleep(d.stream_delay);
			} catch (InterruptedException e) {}
			
		}
	}
	
	//Builds the 
	private void connect(
			String url,
			String con_key,
			String con_sec,
			String acc_token,
			String acc_token_sec) {
		
		OAuthService service = new ServiceBuilder()
			.provider(TwitterApi.class)
			.apiKey(con_key)
			.apiSecret(con_sec)
			.build();
		
		Token accessToken = new Token(acc_token, acc_token_sec);
		OAuthRequest request = new OAuthRequest(Verb.POST, url);
		request.setConnectionKeepAlive(true);
		//Character Set declaration may be required for some connections
		request.setCharset("UTF-8");
		//Content-Length header required for POST to go through
		request.addHeader("Content-Length", Integer.toString(url.length()));
		//request.setConnectTimeout(5, TimeUnit.SECONDS);
		
		service.signRequest(accessToken, request);
		Response response;
		try {
			response = request.send();
		} catch (OAuthException ie) {
			return;
		}
		
		stream = response.getStream();
		inReader = new InputStreamReader(stream);
		buffReader = new BufferedReader(inReader);
		
		d.streaming = true;
		}
	

	//Gets the next tweet and updates the data object to signify a new tweet
	private void getTweet() throws Exception{
		
		//Holds the original string read from the stream
		String source_string = "";
		
		try {
			source_string = buffReader.readLine();
		} catch (IOException ie) {
			System.out.println("Failed to read from stream");
		}

		double[] loc = new double[2];
		try {
			JSONObject json_tweet = new JSONObject(source_string);
			try {
				loc = getLocation(json_tweet);
				if (loc[0] == 0 && loc[1] == 0) {
					throw new Exception("Location 0");
				}
				d.last_tweet = loc;
				d.changed = true;
			} catch (IllegalArgumentException ie) {
				throw new Exception("No location found");
			}
		} catch (JSONException ie) {
			throw new Exception("Could not convert read string into JSONObject");
		}
	}
	
	private double[] getLocation(JSONObject tweet) throws IllegalArgumentException {
		//Gets the geolocated information on a twitter status JSONObject
		
		//Array with longitude, latitude of tweet
		double[] location = new double[2];
	  
		//Attempt to get location from geo-tagged tweets
		try {
			JSONArray geo_tagged = tweet.getJSONObject("coordinates").getJSONArray("coordinates");
			try {
				location[0] = geo_tagged.getDouble(0);
				location[1] = geo_tagged.getDouble(1);
			} catch (JSONException ie) {}
			return location;
			
		} catch (JSONException ie) {}
	  
		//Attempt to get location from tweet place
		try {
			JSONArray place_bb = tweet.getJSONObject("place").getJSONObject("bounding_box").getJSONArray("coordinates");
			JSONArray tweet_place  = place_bb.getJSONArray(0).getJSONArray(0);
			try {
				location[0] = tweet_place.getDouble(0);
				location[1] = tweet_place.getDouble(1);
			} catch (JSONException ie) {}
			return location;
		} catch (JSONException ie) {}
	  
		//If none of the attempts above returned, then we can't get location data from tweet
		throw new IllegalArgumentException("No location data with tweet");
	}
}
